from .oj_download import main as download
from .oj_test import main as test
from .problem import LocalProblem, problem_from_url
from .resolver import OjResolve

__all__ = [
    "LocalProblem",
    "OjResolve",
    "download",
    "problem_from_url",
    "test",
]
